use std::ops::{Add, AddAssign, Div, DivAssign, Index, IndexMut, Mul, MulAssign, Neg, Sub, SubAssign};

use crate::Numeric;

/// Struct representing a dense matrix.
#[repr(transparent)]
#[derive(Copy, Clone, Debug, PartialEq)]
pub struct GenericMatrix<T: Numeric, const ROWS: usize, const COLUMNS: usize> {
    pub data: [[T; COLUMNS]; ROWS],
}

/// Special kind of `GenericMatrix` where both dimensions are equal.
pub type SquareMatrix<T, const DIMENSION: usize> = GenericMatrix<T, DIMENSION, DIMENSION>;

/// Special kind of `GenericMatrix` with just one column.
pub type ColumnVector<T, const ROWS: usize> = GenericMatrix<T, ROWS, 1>;

/// Special kind of `GenericMatrix` with just one row.
pub type RowVector<T, const COLUMNS: usize> = GenericMatrix<T, 1, COLUMNS>;

/// A macro for easier creation of row vectors.
///
/// # Example
///
/// ```rust
/// # use lineal::{RowVector, rvec};
/// let a = rvec![1.0, 2.0, 3.0];
/// // is the same as
/// let b = RowVector {
///     data: [[1.0, 2.0, 3.0]],
/// };
/// ```
#[macro_export]
macro_rules! rvec {
    ($($value:expr),* $(,)?) => {
        {
            RowVector {
                data: [[$( $value, )*]],
            }
        }
    }
}

/// A macro for easier creation of column vectors.
///
/// # Example
///
/// ```rust
/// # use lineal::{ColumnVector, cvec};
/// let a = cvec![1.0, 2.0, 3.0];
/// // is the same as
/// let b = ColumnVector {
///     data: [[1.0], [2.0], [3.0]],
/// };
/// ```
#[macro_export]
macro_rules! cvec {
    ($($value:expr),* $(,)?) => {
        {
            ColumnVector {
                data: [$( [$value], )*],
            }
        }
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> Default for GenericMatrix<T, ROWS, COLUMNS> {
    /// Create a matrix with all cells being zero.
    fn default() -> Self {
        let zero = unsafe { T::whole(0) };
        Self {
            data: [[zero; COLUMNS]; ROWS],
        }
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> GenericMatrix<T, ROWS, COLUMNS> {
    /// Create a transpose of the input matrix.
    ///
    /// # Example
    ///
    /// ```rust
    /// # use lineal::{ColumnVector, RowVector, cvec, rvec};
    /// let a = cvec![1, 2, 3];
    /// let b = rvec![1, 2, 3];
    /// assert_eq!(a.transposed(), b);
    /// ```
    pub fn transposed(&self) -> GenericMatrix<T, COLUMNS, ROWS> {
        let mut result = GenericMatrix::default();
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                result.data[c][r] = self.data[r][c];
            }
        }
        result
    }
}

impl<T: Numeric, const DIMENSION: usize> SquareMatrix<T, DIMENSION> {
    /// Create a square identity matrix.
    ///
    /// In an identity matrix the main diagonal is filled with ones,
    /// while the rest if the cells are zero.
    ///
    /// # Example
    ///
    /// ```rust
    /// # use lineal::{SquareMatrix};
    /// let a: SquareMatrix<f32, 3> = SquareMatrix::identity();
    /// // is the same as
    /// let b: SquareMatrix<f32, 3> = SquareMatrix {
    ///     data: [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    /// };
    /// ```
    pub fn identity() -> Self {
        let mut mat = Self::default();
        let one = unsafe { T::whole(1) };
        for i in 0..DIMENSION {
            mat.data[i][i] = one;
        }
        mat
    }

    /// Create a square matrix with its main diagonal filled with the given values.
    ///
    /// # Example
    ///
    /// ```rust
    /// # use lineal::{SquareMatrix};
    /// let a = SquareMatrix::diagonal(&[1, 2, 3]);
    /// // is the same as
    /// let b = SquareMatrix {
    ///     data: [
    ///         [1, 0, 0],
    ///         [0, 2, 0],
    ///         [0, 0, 3],
    ///     ],
    /// };
    /// ```
    pub fn diagonal(values: &[T; DIMENSION]) -> Self {
        let mut mat = Self::default();
        for i in 0..DIMENSION {
            mat.data[i][i] = values[i];
        }
        mat
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> Index<(usize, usize)> for GenericMatrix<T, ROWS, COLUMNS> {
    type Output = T;

    fn index(&self, index: (usize, usize)) -> &Self::Output {
        let (row, column) = index;
        &self.data[row][column]
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> IndexMut<(usize, usize)> for GenericMatrix<T, ROWS, COLUMNS> {
    fn index_mut(&mut self, index: (usize, usize)) -> &mut Self::Output {
        let (row, column) = index;
        &mut self.data[row][column]
    }
}

impl<T: Numeric + Neg<Output=T>, const ROWS: usize, const COLUMNS: usize> Neg for GenericMatrix<T, ROWS, COLUMNS> {
    type Output = Self;

    fn neg(self) -> Self::Output {
        let mut result = Self::default();
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                result.data[r][c] = -self.data[r][c];
            }
        }
        result
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> Add for GenericMatrix<T, ROWS, COLUMNS> {
    type Output = Self;

    fn add(mut self, rhs: Self) -> Self::Output {
        self += rhs;
        self
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> AddAssign for GenericMatrix<T, ROWS, COLUMNS> {
    fn add_assign(&mut self, rhs: Self) {
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                self.data[r][c] += rhs.data[r][c];
            }
        }
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> Sub for GenericMatrix<T, ROWS, COLUMNS> {
    type Output = Self;

    fn sub(mut self, rhs: Self) -> Self::Output {
        self -= rhs;
        self
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> SubAssign for GenericMatrix<T, ROWS, COLUMNS> {
    fn sub_assign(&mut self, rhs: Self) {
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                self.data[r][c] -= rhs.data[r][c];
            }
        }
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> Mul<T> for GenericMatrix<T, ROWS, COLUMNS> {
    type Output = Self;

    fn mul(mut self, rhs: T) -> Self::Output {
        self *= rhs;
        self
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> MulAssign<T> for GenericMatrix<T, ROWS, COLUMNS> {
    fn mul_assign(&mut self, rhs: T) {
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                self.data[r][c] *= rhs;
            }
        }
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> Div<T> for GenericMatrix<T, ROWS, COLUMNS> {
    type Output = Self;

    fn div(mut self, rhs: T) -> Self::Output {
        self /= rhs;
        self
    }
}

impl<T: Numeric, const ROWS: usize, const COLUMNS: usize> DivAssign<T> for GenericMatrix<T, ROWS, COLUMNS> {
    fn div_assign(&mut self, rhs: T) {
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                self.data[r][c] /= rhs;
            }
        }
    }
}

impl<T: Numeric, const ROWS: usize, const COMMON: usize, const COLUMNS: usize> Mul<GenericMatrix<T, COMMON, COLUMNS>> for GenericMatrix<T, ROWS, COMMON> {
    type Output = GenericMatrix<T, ROWS, COLUMNS>;

    fn mul(self, rhs: GenericMatrix<T, COMMON, COLUMNS>) -> Self::Output {
        let mut result = Self::Output::default();
        for i in 0..ROWS {
            for j in 0..COLUMNS {
                for k in 0..COMMON {
                    result.data[i][j] += self.data[i][k] * rhs.data[k][j];
                }
            }
        }
        result
    }
}

#[cfg(test)]
mod tests {
    use crate::{ColumnVector, Complex, cplx, GenericMatrix, RowVector, SquareMatrix};

    #[test]
    fn identity_matrix() {
        let mat: SquareMatrix<f64, 3> = SquareMatrix::identity();
        let expected = SquareMatrix {
            data: [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
        };

        assert_eq!(mat, expected);
    }

    #[test]
    fn diagonal_matrix() {
        let mat: SquareMatrix<f32, 3> = SquareMatrix::diagonal(&[1.0, 2.0, 3.0]);
        let expected = SquareMatrix {
            data: [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]],
        };

        assert_eq!(mat, expected);
    }

    #[test]
    #[should_panic = "index out of bounds"]
    fn out_of_bounds_access() {
        let mat: SquareMatrix<f32, 1> = SquareMatrix::identity();
        mat[(1, 0)];
    }

    #[test]
    fn transposing_matrix() {
        let mat = GenericMatrix {
            data: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]],
        }.transposed();
        let expected = GenericMatrix {
            data: [[1.0, 3.0, 5.0], [2.0, 4.0, 6.0]],
        };

        assert_eq!(mat, expected);
    }

    #[test]
    fn negating_matrix() {
        let mat = -GenericMatrix {
            data: [[1, -1], [2, -2]],
        };
        let expected = GenericMatrix {
            data: [[-1, 1], [-2, 2]],
        };

        assert_eq!(mat, expected);
    }

    #[test]
    fn dot_product_of_vectors() {
        let a = rvec![1, 3, -5];
        let b = cvec![4, -2, -1];
        let product = a * b;

        assert_eq!(product[(0, 0)], 3);
    }

    #[test]
    fn complex_matrix_multiplication() {
        let a = SquareMatrix {
            data: [
                [cplx!(2.0, 1.0), cplx!(i = 5.0,)],
                [cplx!(3.0), cplx!(3.0, -4.0)],
            ],
        };
        let b = SquareMatrix {
            data: [
                [cplx!(1.0, -1.0), cplx!(4.0, 2.0)],
                [cplx!(1.0, -6.0), cplx!(3.0)],
            ],
        };
        let mat = a * b;
        let expected = SquareMatrix {
            data: [
                [cplx!(33.0, 4.0), cplx!(6.0, 23.0)],
                [cplx!(-18.0, -25.0), cplx!(21.0, -6.0)],
            ],
        };

        assert_eq!(mat, expected);
    }
}
